/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#include "Coupled.h"
#include "Atomic.h"
#include "ODEVS_Exception.h"
#include "ClassAdmin.h"
#include <assert.h>
#include <algorithm>
#include <float.h>

using namespace ODEVS;
 
IMPLEMENT_CLASS1(Coupled, "Coupled", 20060806, Model);

void Coupled::AddModel(Model* md) 
{ 
	if(find(M.begin(), M.end(), md)==M.end()){
		M.push_back(md); 
		md->Parent = this;
	}
}

Coupled::~Coupled()
{
	while(M.empty()==false)
	{
		Model* md = M.front();
		M.pop_front();
		delete md;
	}
}

Model* Coupled::GetModel(const string& name) const
{
	for(deque<Model*>::size_type i = 0; i < M.size(); i++)
		if(M[i]->Name == name)
			return M[i];

	return NULL;
}

void Coupled::Update_tL_tN()
{
	sort(Schedule.begin(), Schedule.end(), lt_MtN());
	if(Schedule.empty() == false){
		t_Next = Schedule.front()->TimeNext();
		for(deque<Model*>::size_type i=0; i<M.size(); i++)
		{
			Time tl = M[i]->TimeLast();
			if( t_Last < tl) t_Last = tl; // t_Last=max(...);
		}
	}
}

//-- external input coupling
void Coupled::Couple(Model* smd, InputPort spt, Model* dmd, InputPort dpt)
{
	assert(smd == this && dmd != this);
	C.insert(std::pair<ModelPort, ModelPort>(ModelPort(this,spt), ModelPort(dmd, dpt)) );
}

//-- internal coupling
void Coupled::Couple(Model* smd, OutputPort spt, Model* dmd, InputPort dpt)
{
	assert(smd != this && dmd != this);
	C.insert(std::pair<ModelPort, ModelPort>(ModelPort(smd, spt), ModelPort(dmd, dpt)));
}

//-- external output coupling
void Coupled::Couple(Model* smd, OutputPort spt, Model* dmd, OutputPort dpt)
{
	assert(smd != this && dmd == this);
	C.insert(std::pair<ModelPort, ModelPort>(ModelPort(smd, spt), ModelPort(this, dpt)));
}

string Coupled::Get_q(bool remaining) const
{
	string str="("+Name+":";
	for(deque<Model*>::size_type i = 0; i < M.size(); i++)
	{
		if(i>0) str+=",";
		str+=M[i]->Get_q(remaining);
	}

	//char tmp[100];
	//if(TimeNext() == DBL_MAX) // passive
	//	if(remaining)
	//		sprintf(tmp, ",inf");
	//	else // elapsed time
	//		sprintf(tmp, ",%lf", TimeElapsed());
	//else // active 
	//	sprintf(tmp, ",%lf", (remaining)? TimeRemaining(): TimeElapsed() );
	//str+=tmp;
	str+=")";
	return str;
}

string HierName(const Coupled* p)
{
	string str;
	if(p->Parent)
		str = HierName(p->Parent)+".";
	str += p->Name;
	return str;
}

void Coupled::PrintCouplings() const
{
	cout << "Inside of " << HierName(this) << endl;
	for(CoupleMap::const_iterator pos = C.begin(); pos != C.end(); pos++)
	{
		ModelPort fmp = pos->first;
		ModelPort smp = pos->second;
		cout << fmp.model->Name <<"."<<fmp.port<<"-->"<<smp.model->Name<<"."<<smp.port << endl;
	}
	for(deque<Model*>::size_type i = 0; i < M.size(); i++){
		if(M[i]->IsKindOf(ODEVS_RT(Coupled)))
			((Coupled*)M[i])->PrintCouplings();
	}
}

//----------------------------------------------------------------------------------
void Coupled::when_receive_init()
{
	Schedule.clear();
	for(deque<Model*>::size_type i = 0; i < M.size(); i++){
		M[i]->when_receive_init();
		Schedule.push_back(M[i]);
	}

	t_Last = 0;
	Update_tL_tN();
}

void Coupled::when_receive_star(bool print_event)
{
	if(Schedule.empty() == false && IsEqual(t_Next, t_Current, 1E-5) ) {
		if(Schedule.empty() == false)
		{
			Model* first = Schedule.front();
			first->when_receive_star(print_event);
			Update_tL_tN();
		}
	}else
		THROW_ODEVS_EXCEPTION("Schedule Violation in Coupled!");
}

void Coupled::when_receive_phi()
{
	if(t_Last <= t_Current && t_Current <= t_Next) {
		for(deque<Model*>::size_type i = 0; i < M.size(); i++)
			M[i]->when_receive_phi();
	}else
		THROW_ODEVS_EXCEPTION("Schedule Violation in Coupled!");
}
void Coupled::when_receive_x(const PortValue& x,bool print_event)
{
	if(t_Last <= t_Current && t_Current <= t_Next) {
		//-- transmit x to influencee along EIC
		for(CoupleMap::iterator pos = C.lower_bound(ModelPort(this,x.port)); 
			pos != C.upper_bound(ModelPort(this,x.port)); pos++)
		{
			ModelPort Influencee = pos->second;
			PortValue pv(Influencee.port, x.value);
			Influencee.model->when_receive_x(pv,print_event);
			Update_tL_tN();
		}
	}else
		THROW_ODEVS_EXCEPTION("Schedule Violation in Coupled!");
}

void Coupled::GetInfluencees_x(InputPort x, deque<pair<Atomic*, InputPort> >& M_X) const
{
	ModelPort m_p((Model*)this, x);

	for(CoupleMap::const_iterator it = C.lower_bound(m_p); it!= C.upper_bound(m_p); it++)
	{
		if(it->second.model->IsKindOf(ODEVS_RT(Atomic))){
			M_X.push_back(pair<Atomic*, InputPort>((Atomic*)it->second.model, InputPort(it->second.port)));
		}else if(it->second.model->IsKindOf(ODEVS_RT(Coupled)))
			((Coupled*)it->second.model)->GetInfluencees_x(it->second.port, M_X);
		else
			THROW_ODEVS_EXCEPTION("Invalid Model!");
	}
}

void Coupled::when_receive_y(Model& child, const PortValue& y,bool print_event)
{
	ModelPort mp(&child, y.port);
	for(CoupleMap::iterator pos = C.lower_bound(mp); 
		pos != C.upper_bound(mp); pos++)
	{//-- transmit y to influencee along ITC
		ModelPort Influencee = pos->second;
		PortValue pv(Influencee.port, y.value);
		if(Influencee.model == this ) 
		{// external output coupling
			if(Parent)
				Parent->when_receive_y(*this, pv,print_event);
		}
		else { // internal coupling
			Influencee.model->when_receive_x(pv,print_event);
			Update_tL_tN();
		}
	}
}


void Coupled::GetInfluencees_y(Model* child, OutputPort y, deque<pair<Atomic*, InputPort> >& M_X) const
{
	ModelPort m_p(child, y);

	for(CoupleMap::const_iterator it = C.lower_bound(m_p); it!= C.upper_bound(m_p); it++)
	{
		if(it->second.model->IsKindOf(ODEVS_RT(Atomic))){ // internal coupling
			M_X.push_back(pair<Atomic*, InputPort>((Atomic*)it->second.model, InputPort(it->second.port)));
		}else if(it->second.model->IsKindOf(ODEVS_RT(Coupled)))
		{
			if((const Coupled*)it->second.model == this){ // external output coupling
				if(Parent)
					Parent->GetInfluencees_y((Model*)this, it->second.port, M_X);
			}
			else // internal coupling 
				((Coupled*)it->second.model)->GetInfluencees_x(it->second.port, M_X);
		}
		else
			THROW_ODEVS_EXCEPTION("Invalid Model!");
	}
}