/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#pragma once
 
#include <string>

#include "PortValue.h"
#include "ODEVS_Export.h"
#include "DeclareClass.h"

using namespace std;

namespace ODEVS {
typedef double Time;
typedef double TimeSpan ;
 
///--------A class having a name ------------------
class ODEVS_EXP Named
{
public:
	Named(const string& name):Name(name){}
	string Name;
};//-----------------------------------------------
	 

class Coupled; // see Coupled.h for details.

/// base class of DEVS models
class ODEVS_EXP Model: public Named
{
public:	
	/// get last-schedule-update time
	Time TimeLast() const { return t_Last; }
	/// get next-schedule time
	Time TimeNext() const { return t_Next; }
	/// get current time;
	static Time TimeCurrent() { return t_Current; }
	/// get reamining time to next schedule
	Time TimeRemaining() const { return t_Next - t_Current; }
	/// get elapsed time since last schedule
	Time TimeElapsed() const { return t_Current - t_Last; }
	/// virtual destructor 
	virtual ~Model() {}

	/* return string of (s,ta(s),t_r) if remaining==true
	   return string of (s,ta(s),t_e) otherwise */
	virtual string Get_q(bool remaining=true) const {return string(); }

	// parent pointer
	Coupled*  Parent;
protected:
	// protected constructor => abstracted class
	Model(const string& name);

	//--- functions for abstract simulation algorithm ---
	/// function called when simulation initialized
	virtual void when_receive_init()=0;
	/// function called when schedule-time reached,(print_event for debugging)
	virtual void when_receive_star(bool print_event)=0; 
	/// function called when continuous time reached
	virtual void when_receive_phi()=0;  
	/// function called when an external input "x" comes 
	virtual void when_receive_x(const PortValue& x, bool print_event)=0; 
	//----------------------------------------------------

private:
	//------------ discrete updating times ---------------
	Time t_Last; // last-schedule-update time
	Time t_Next; // next-schedule time
	//----------------------------------------------------
friend class Coupled;
friend class Atomic;
	//--------------- continuously times -----------------
	static Time t_Current; // current time
	//----------------------------------------------------
friend class SRTEngine;

	DECLARE_CLASS();

};


/* A comparable structure.  
   This will be used for sorting models by their next-event schedules */
struct lt_MtN
{
  bool operator()(const Model* s1, const Model* s2) const
  {
    return (s1->TimeNext() < s2->TimeNext()) ;
  }
};//-------------------------------------------------------


/* A comparable structure.  
   This will be used for sorting models by their names */
struct lt_MName
{
  bool operator()(const Model* s1, const Model* s2) const
  {
    return (s1->Name < s2->Name) ;
  }
};//-------------------------------------------------------

//-- check two time values a and b are equal in terms of tolerance tol
ODEVS_EXP bool IsEqual(Time a, Time b, double tol=1E-3);

// get the string of hierarchical name from child to under
ODEVS_EXP string GetHierName(const Coupled* under, const Model* child) ;

} // end of namespace of ODEVS 
