/*************************************************************************
 * ODEVS-C++: Open Source of DEVS in C++                                 *
 * Copyright (C) 2005-2006  Moon Ho Hwang <moonho.hwang@gmail.com>       *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2.1 of the License, or (at your option) any later version.    *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, *
 * USA.                                                                  *
 *************************************************************************/
#pragma once

#include "PortValue.h"
#include "Model.h"
#include "ODEVS_Export.h"

#include <list>
#include <float.h>

/*-------------------  Functionalities: -------------------------------
   (0) A simulation engine which has its own thread. 
   (1) Its simulation run can perform when "Run" function called.
   (2) The simuation speed is controllable according to "m_time_scale" 
       variable.
   (3) Pause: stop a simulation run.
       Reset: Pause and initialize the current time.
	   Rerun: reset and run.
   (4) Considering the schedule of simulation models
       (4.1) pure virtual simulation
	   (4.2) soft real-time simulation
   (5) Step-by-Schedule mode (as well as run-through mode)
   (6) Handling the external input event coming from outside of 
       the simulation environment
   (7) simulation ending time
   (9) animation of analytical continuous change
   (10) printing menu 
   (11) Callback function for Inject handling
-----------------------------------------------------------------------*/
namespace ODEVS {
#pragma warning(disable:4251)

class ODEVS_EXP SRTEngine
{  
	std::list<PortValue> m_ExtX;  

	void TransmitExtEvent();

	enum STEP_MODE { GO_THROUGH, STEP_BY_ONE_SCHEDULE };

	STEP_MODE m_StepMode;

	/// execution function of simulation model at schedule time
	void ScheduleExecution(); 

	/// hightest simulator  
	Model& m_model;

	/// Previous Clock Time
    unsigned long m_PreviousClockTime; //-- 
  
	/// Flag of Pause requested
	bool m_bPauseReq;  

	/// Flag of Pause compled
	bool m_bPauseComp;  

	/* Time Scale:  
	 	scale=0.5      => speed=RT/2.0; (2 times slower)
	 	scale=1        => speed=RT; (fast as real time if we can)
	 	scale=2.0      => speed=2*RT; (two times faster if we can)
	 	scale=INFINITE => speed=INFT; (pure discrete event simulation)
	*/
    double m_time_scale;

    /// Flag of Exit requested
 	bool m_bExitReq;

	/// Flag of Exit completed
	bool m_bExitComp;

	////
	bool m_bigmouth;

	Time m_ending_time; //

	string m_menu;

	TimeSpan m_animation_interval;

public:
	enum PrintStateMode {P_NONE, P_Q_t_remaining, P_Q_t_elapsed} ;
	typedef PortValue (*CallBack)();
private:
	PrintStateMode print_state_change; //

	bool b_acs; // flag for analytical continuous simulation

	void AnimationExecution();

	CallBack call_back_method;
public:
	/* Model& modl: DEVS Model to be simulated;
	   CallBack cbf: function pointer whose form is PortValue (*CallBack)() which
                     is supposed to return an input event to be used for injection;
	   Time ending_t: simulation termination time;
	*/
	SRTEngine(Model& modl, CallBack cbf=NULL, Time ending_t = DBL_MAX);

	/// Destructor
    virtual ~SRTEngine() { KillLoop(); }
	
	/// Inject a portvalue into the simulator's input bag
    void Inject(PortValue x)
    {
        if (m_bPauseComp == true)
            StampClockTime(); // stemp the invoking time
		m_ExtX.push_front(x); // add x to m_ExtX buffer
		m_bPauseReq = false; // awacken 
    }
	
	/// Set Runing Flag on called by Outside 
    void Step()
    {
        if (m_bPauseReq == false && m_StepMode == STEP_BY_ONE_SCHEDULE)
            return;

        m_StepMode = STEP_BY_ONE_SCHEDULE;
        StampClockTime();
        m_bPauseReq = false;
    }

	/// Initialize control Variables
	void Init()
	{
		m_model.when_receive_init();  
		m_bPauseReq = true;  
		m_bPauseComp = true; 
		m_bExitReq = false;  
		m_bExitComp = false;  
		Model::t_Current = 0.0;
	}

	/// Starter for a simulation run
	void Run()
	{	 
		if(m_bPauseReq == true){ 
			m_StepMode = GO_THROUGH; 
			StampClockTime();
			m_bPauseReq = false;  
		}
	}
	/// pause the simulation run. Called by a non-simulation thread.
    void Pause() ;

	/// Reinitialize Model
	void Reset()
    {
        Pause();  
		Model::t_Current = 0.0;
		m_model.when_receive_init();  
    }

    /// Reset models
    void Rerun() { Reset(); Run(); }
	 
	/// Simulation Loop which is supposed to be called by a simulation "THREAD"
    void Loop( );
	void KillLoop();

	/// Get the time scale 
    double GetTimeScale() const { return m_time_scale; }
	/// Set the time scale
    void SetTimeScale(double ts);

	Time GetEndingTime() const { return m_ending_time; }
	void SetEndingTime(Time et) { m_ending_time = et; }

	void Set_dtmode(PrintStateMode flag) { print_state_change = flag; } 
	PrintStateMode Get_dtmode() const { return print_state_change; }

	void SetAnimationFlag(bool flag) { b_acs = flag; } 
	bool GetAnimationFlag() const { return b_acs; }

	void SetAnimationInterval(TimeSpan ai) { m_animation_interval = ai; }
	TimeSpan GetAnimationInterval() const { return m_animation_interval; }

	void RunConsoleMenu();

	//-- print all settins 
	void PrintSettings() const;
private: 
	/// Remember the clcok time when getting start of a run.
	void StampClockTime();
};
 
}// end of namespace of ODEVS